@@ -1,7 +1,7 @@
 // Three-lobed Lanczos, the most common choice.
 // Note that if you change this, the accuracy for LANCZOS_TABLE_SIZE
 // needs to be recomputed.
-#define LANCZOS_RADIUS 3.0
+#define LANCZOS_RADIUS 3.0f
 
 #include <epoxy/gl.h>
 #include <assert.h>
@@ -27,12 +27,6 @@ namespace movit {
 
 namespace {
 
-template<class T>
-struct Tap {
-	T weight;
-	T pos;
-};
-
 float sinc(float x)
 {
 	if (fabs(x) < 1e-6) {
@@ -96,7 +90,7 @@ float lanczos_weight_cached(float x)
 		return 0.0f;
 	}
 	float table_pos = x * (LANCZOS_TABLE_SIZE / LANCZOS_RADIUS);
-	int table_pos_int = int(table_pos);  // Truncate towards zero.
+	unsigned table_pos_int = int(table_pos);  // Truncate towards zero.
 	float table_pos_frac = table_pos - table_pos_int;
 	assert(table_pos < LANCZOS_TABLE_SIZE + 2);
 	return lanczos_table[table_pos_int] +
@@ -115,7 +109,7 @@ unsigned gcd(unsigned a, unsigned b)
 }
 
 template<class DestFloat>
-unsigned combine_samples(const Tap<float> *src, Tap<DestFloat> *dst, float num_subtexels, float inv_num_subtexels, unsigned num_src_samples, unsigned max_samples_saved)
+unsigned combine_samples(const Tap<float> *src, Tap<DestFloat> *dst, float num_subtexels, float inv_num_subtexels, unsigned num_src_samples, unsigned max_samples_saved, float pos1_pos2_diff, float inv_pos1_pos2_diff)
 {
 	// Cut off near-zero values at both sides.
 	unsigned num_samples_saved = 0;
@@ -163,7 +157,7 @@ unsigned combine_samples(const Tap<float> *src, Tap<DestFloat> *dst, float num_s
 
 		DestFloat pos, total_weight;
 		float sum_sq_error;
-		combine_two_samples(w1, w2, pos1, pos2, num_subtexels, inv_num_subtexels, &pos, &total_weight, &sum_sq_error);
+		combine_two_samples(w1, w2, pos1, pos1_pos2_diff, inv_pos1_pos2_diff, num_subtexels, inv_num_subtexels, &pos, &total_weight, &sum_sq_error);
 
 		// If the interpolation error is larger than that of about sqrt(2) of
 		// a level at 8-bit precision, don't combine. (You'd think 1.0 was enough,
@@ -216,15 +210,18 @@ unsigned combine_many_samples(const Tap<float> *weights, unsigned src_size, unsi
 {
 	float num_subtexels = src_size / movit_texel_subpixel_precision;
 	float inv_num_subtexels = movit_texel_subpixel_precision / src_size;
+	float pos1_pos2_diff = 1.0f / src_size;
+	float inv_pos1_pos2_diff = src_size;
 
 	unsigned max_samples_saved = UINT_MAX;
 	for (unsigned y = 0; y < dst_samples && max_samples_saved > 0; ++y) {
-		unsigned num_samples_saved = combine_samples<DestFloat>(weights + y * src_samples, NULL, num_subtexels, inv_num_subtexels, src_samples, max_samples_saved);
+		unsigned num_samples_saved = combine_samples<DestFloat>(weights + y * src_samples, NULL, num_subtexels, inv_num_subtexels, src_samples, max_samples_saved, pos1_pos2_diff, inv_pos1_pos2_diff);
 		max_samples_saved = min(max_samples_saved, num_samples_saved);
 	}
 
 	// Now that we know the right width, actually combine the samples.
 	unsigned src_bilinear_samples = src_samples - max_samples_saved;
+	if (*bilinear_weights != NULL) delete[] *bilinear_weights;
 	*bilinear_weights = new Tap<DestFloat>[dst_samples * src_bilinear_samples];
 	for (unsigned y = 0; y < dst_samples; ++y) {
 		Tap<DestFloat> *bilinear_weights_ptr = *bilinear_weights + y * src_bilinear_samples;
@@ -234,7 +231,9 @@ unsigned combine_many_samples(const Tap<float> *weights, unsigned src_size, unsi
 			num_subtexels,
 			inv_num_subtexels,
 			src_samples,
-			max_samples_saved);
+			max_samples_saved,
+			pos1_pos2_diff,
+			inv_pos1_pos2_diff);
 		assert(num_samples_saved == max_samples_saved);
 		normalize_sum(bilinear_weights_ptr, src_bilinear_samples);
 	}
@@ -255,10 +254,10 @@ double compute_sum_sq_error(const Tap<float>* weights, unsigned num_weights,
 	// Find the effective range of the bilinear-optimized kernel.
 	// Due to rounding of the positions, this is not necessarily the same
 	// as the intended range (ie., the range of the original weights).
-	int lower_pos = int(floor(to_fp32(bilinear_weights[0].pos) * size - 0.5));
-	int upper_pos = int(ceil(to_fp32(bilinear_weights[num_bilinear_weights - 1].pos) * size - 0.5)) + 2;
-	lower_pos = min<int>(lower_pos, lrintf(weights[0].pos * size - 0.5));
-	upper_pos = max<int>(upper_pos, lrintf(weights[num_weights - 1].pos * size - 0.5) + 1);
+	int lower_pos = int(floor(to_fp32(bilinear_weights[0].pos) * size - 0.5f));
+	int upper_pos = int(ceil(to_fp32(bilinear_weights[num_bilinear_weights - 1].pos) * size - 0.5f)) + 2;
+	lower_pos = min<int>(lower_pos, lrintf(weights[0].pos * size - 0.5f));
+	upper_pos = max<int>(upper_pos, lrintf(weights[num_weights - 1].pos * size - 0.5f) + 1);
 
 	float* effective_weights = new float[upper_pos - lower_pos];
 	for (int i = 0; i < upper_pos - lower_pos; ++i) {
@@ -277,7 +276,7 @@ double compute_sum_sq_error(const Tap<float>* weights, unsigned num_weights,
 		assert(x0 < upper_pos - lower_pos);
 		assert(x1 < upper_pos - lower_pos);
 
-		effective_weights[x0] += to_fp32(bilinear_weights[i].weight) * (1.0 - f);
+		effective_weights[x0] += to_fp32(bilinear_weights[i].weight) * (1.0f - f);
 		effective_weights[x1] += to_fp32(bilinear_weights[i].weight) * f;
 	}
 
@@ -302,7 +301,8 @@ double compute_sum_sq_error(const Tap<float>* weights, unsigned num_weights,
 }  // namespace
 
 ResampleEffect::ResampleEffect()
-	: input_width(1280),
+	: owns_effects(true),
+	  input_width(1280),
 	  input_height(720),
 	  offset_x(0.0f), offset_y(0.0f),
 	  zoom_x(1.0f), zoom_y(1.0f),
@@ -320,6 +320,14 @@ ResampleEffect::ResampleEffect()
 	update_size();
 }
 
+ResampleEffect::~ResampleEffect()
+{
+	if (owns_effects) {
+		delete hpass;
+		delete vpass;
+	}
+}
+
 void ResampleEffect::rewrite_graph(EffectChain *graph, Node *self)
 {
 	Node *hpass_node = graph->add_node(hpass);
@@ -328,6 +336,7 @@ void ResampleEffect::rewrite_graph(EffectChain *graph, Node *self)
 	graph->replace_receiver(self, hpass_node);
 	graph->replace_sender(self, vpass_node);
 	self->disabled = true;
+	owns_effects = false;
 } 
 
 // We get this information forwarded from the first blur pass,
@@ -506,12 +515,71 @@ void SingleResamplePassEffect::update_texture(GLuint glsl_program_num, const str
 		assert(false);
 	}
 
+	ScalingWeights weights = calculate_scaling_weights(src_size, dst_size, zoom, offset);
+	src_bilinear_samples = weights.src_bilinear_samples;
+	num_loops = weights.num_loops;
+	slice_height = 1.0f / weights.num_loops;
+
+	// Encode as a two-component texture. Note the GL_REPEAT.
+	glActiveTexture(GL_TEXTURE0 + *sampler_num);
+	check_error();
+	glBindTexture(GL_TEXTURE_2D, texnum);
+	check_error();
+	if (last_texture_width == -1) {
+		// Need to set this state the first time.
+		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+		check_error();
+		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
+		check_error();
+		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
+		check_error();
+	}
+
+	GLenum type, internal_format;
+	void *pixels;
+	assert((weights.bilinear_weights_fp16 == NULL) != (weights.bilinear_weights_fp32 == NULL));
+	if (weights.bilinear_weights_fp32 != NULL) {
+		type = GL_FLOAT;
+		internal_format = GL_RG32F;
+		pixels = weights.bilinear_weights_fp32;
+	} else {
+		type = GL_HALF_FLOAT;
+		internal_format = GL_RG16F;
+		pixels = weights.bilinear_weights_fp16;
+	}
+
+	if (int(weights.src_bilinear_samples) == last_texture_width &&
+	    int(weights.dst_samples) == last_texture_height &&
+	    internal_format == last_texture_internal_format) {
+		// Texture dimensions and type are unchanged; it is more efficient
+		// to just update it rather than making an entirely new texture.
+		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, weights.src_bilinear_samples, weights.dst_samples, GL_RG, type, pixels);
+	} else {
+		glTexImage2D(GL_TEXTURE_2D, 0, internal_format, weights.src_bilinear_samples, weights.dst_samples, 0, GL_RG, type, pixels);
+		last_texture_width = weights.src_bilinear_samples;
+		last_texture_height = weights.dst_samples;
+		last_texture_internal_format = internal_format;
+	}
+	check_error();
+
+	delete[] weights.bilinear_weights_fp16;
+	delete[] weights.bilinear_weights_fp32;
+}
+
+ScalingWeights calculate_scaling_weights(unsigned src_size, unsigned dst_size, float zoom, float offset)
+{
+	if (!lanczos_table_init_done) {
+		// Only needed if run from outside ResampleEffect.
+		init_lanczos_table();
+	}
+
 	// For many resamplings (e.g. 640 -> 1280), we will end up with the same
 	// set of samples over and over again in a loop. Thus, we can compute only
 	// the first such loop, and then ask the card to repeat the texture for us.
 	// This is both easier on the texture cache and lowers our CPU cost for
 	// generating the kernel somewhat.
 	float scaling_factor;
+	int num_loops;
 	if (fabs(zoom - 1.0f) < 1e-6) {
 		num_loops = gcd(src_size, dst_size);
 		scaling_factor = float(dst_size) / float(src_size);
@@ -524,7 +592,6 @@ void SingleResamplePassEffect::update_texture(GLuint glsl_program_num, const str
 		num_loops = 1;
 		scaling_factor = zoom * float(dst_size) / float(src_size);
 	}
-	slice_height = 1.0f / num_loops;
 	unsigned dst_samples = dst_size / num_loops;
 
 	// Sample the kernel in the right place. A diagram with a triangular kernel
@@ -589,11 +656,12 @@ void SingleResamplePassEffect::update_texture(GLuint glsl_program_num, const str
 		int base_src_y = lrintf(center_src_y);
 
 		// Now sample <int_radius> pixels on each side around that point.
+		float inv_src_size = 1.0 / float(src_size);
 		for (int i = 0; i < src_samples; ++i) {
 			int src_y = base_src_y + i - int_radius;
 			float weight = lanczos_weight_cached(radius_scaling_factor * (src_y - center_src_y - subpixel_offset));
 			weights[y * src_samples + i].weight = weight * radius_scaling_factor;
-			weights[y * src_samples + i].pos = (src_y + 0.5) / float(src_size);
+			weights[y * src_samples + i].pos = (src_y + 0.5f) * inv_src_size;
 		}
 	}
 
@@ -602,10 +670,9 @@ void SingleResamplePassEffect::update_texture(GLuint glsl_program_num, const str
 	// Our tolerance level for total error is a bit higher than the one for invididual
 	// samples, since one would assume overall errors in the shape don't matter as much.
 	const float max_error = 2.0f / (255.0f * 255.0f);
-	Tap<fp16_int_t> *bilinear_weights_fp16;
-	src_bilinear_samples = combine_many_samples(weights, src_size, src_samples, dst_samples, &bilinear_weights_fp16);
+	Tap<fp16_int_t> *bilinear_weights_fp16 = NULL;
+	int src_bilinear_samples = combine_many_samples(weights, src_size, src_samples, dst_samples, &bilinear_weights_fp16);
 	Tap<float> *bilinear_weights_fp32 = NULL;
-	bool fallback_to_fp32 = false;
 	double max_sum_sq_error_fp16 = 0.0;
 	for (unsigned y = 0; y < dst_samples; ++y) {
 		double sum_sq_error_fp16 = compute_sum_sq_error(
@@ -619,54 +686,20 @@ void SingleResamplePassEffect::update_texture(GLuint glsl_program_num, const str
 	}
 
 	if (max_sum_sq_error_fp16 > max_error) {
-		fallback_to_fp32 = true;
+		delete[] bilinear_weights_fp16;
+		bilinear_weights_fp16 = NULL;
 		src_bilinear_samples = combine_many_samples(weights, src_size, src_samples, dst_samples, &bilinear_weights_fp32);
 	}
 
-	// Encode as a two-component texture. Note the GL_REPEAT.
-	glActiveTexture(GL_TEXTURE0 + *sampler_num);
-	check_error();
-	glBindTexture(GL_TEXTURE_2D, texnum);
-	check_error();
-	if (last_texture_width == -1) {
-		// Need to set this state the first time.
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
-		check_error();
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-		check_error();
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-		check_error();
-	}
-
-	GLenum type, internal_format;
-	void *pixels;
-	if (fallback_to_fp32) {
-		type = GL_FLOAT;
-		internal_format = GL_RG32F;
-		pixels = bilinear_weights_fp32;
-	} else {
-		type = GL_HALF_FLOAT;
-		internal_format = GL_RG16F;
-		pixels = bilinear_weights_fp16;
-	}
-
-	if (int(src_bilinear_samples) == last_texture_width &&
-	    int(dst_samples) == last_texture_height &&
-	    internal_format == last_texture_internal_format) {
-		// Texture dimensions and type are unchanged; it is more efficient
-		// to just update it rather than making an entirely new texture.
-		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, src_bilinear_samples, dst_samples, GL_RG, type, pixels);
-	} else {
-		glTexImage2D(GL_TEXTURE_2D, 0, internal_format, src_bilinear_samples, dst_samples, 0, GL_RG, type, pixels);
-		last_texture_width = src_bilinear_samples;
-		last_texture_height = dst_samples;
-		last_texture_internal_format = internal_format;
-	}
-	check_error();
-
 	delete[] weights;
-	delete[] bilinear_weights_fp16;
-	delete[] bilinear_weights_fp32;
+
+	ScalingWeights ret;
+	ret.src_bilinear_samples = src_bilinear_samples;
+	ret.dst_samples = dst_samples;
+	ret.num_loops = num_loops;
+	ret.bilinear_weights_fp16 = bilinear_weights_fp16;
+	ret.bilinear_weights_fp32 = bilinear_weights_fp32;
+	return ret;
 }
 
 void SingleResamplePassEffect::set_gl_state(GLuint glsl_program_num, const string &prefix, unsigned *sampler_num)
