@@ -19,10 +19,13 @@ namespace movit {
 YCbCrInput::YCbCrInput(const ImageFormat &image_format,
                        const YCbCrFormat &ycbcr_format,
                        unsigned width, unsigned height,
-                       YCbCrInputSplitting ycbcr_input_splitting)
+                       YCbCrInputSplitting ycbcr_input_splitting,
+                       GLenum type)
 	: image_format(image_format),
 	  ycbcr_format(ycbcr_format),
 	  ycbcr_input_splitting(ycbcr_input_splitting),
+	  needs_mipmaps(false),
+	  type(type),
 	  width(width),
 	  height(height),
 	  resource_pool(NULL)
@@ -38,7 +41,11 @@ YCbCrInput::YCbCrInput(const ImageFormat &image_format,
 
 	register_uniform_sampler2d("tex_y", &uniform_tex_y);
 
-	if (ycbcr_input_splitting == YCBCR_INPUT_SPLIT_Y_AND_CBCR) {
+	if (ycbcr_input_splitting == YCBCR_INPUT_INTERLEAVED) {
+		num_channels = 1;
+		assert(ycbcr_format.chroma_subsampling_x == 1);
+		assert(ycbcr_format.chroma_subsampling_y == 1);
+	} else if (ycbcr_input_splitting == YCBCR_INPUT_SPLIT_Y_AND_CBCR) {
 		num_channels = 2;
 		register_uniform_sampler2d("tex_cbcr", &uniform_tex_cb);
 	} else {
@@ -47,6 +54,12 @@ YCbCrInput::YCbCrInput(const ImageFormat &image_format,
 		register_uniform_sampler2d("tex_cb", &uniform_tex_cb);
 		register_uniform_sampler2d("tex_cr", &uniform_tex_cr);
 	}
+
+	register_int("needs_mipmaps", &needs_mipmaps);
+	register_uniform_mat3("inv_ycbcr_matrix", &uniform_ycbcr_matrix);
+	register_uniform_vec3("offset", uniform_offset);
+	register_uniform_vec2("cb_offset", (float *)&uniform_cb_offset);
+	register_uniform_vec2("cr_offset", (float *)&uniform_cr_offset);
 }
 
 YCbCrInput::~YCbCrInput()
@@ -58,25 +71,59 @@ YCbCrInput::~YCbCrInput()
 
 void YCbCrInput::set_gl_state(GLuint glsl_program_num, const string& prefix, unsigned *sampler_num)
 {
+	compute_ycbcr_matrix(ycbcr_format, uniform_offset, &uniform_ycbcr_matrix, type);
+
+	uniform_cb_offset.x = compute_chroma_offset(
+		ycbcr_format.cb_x_position, ycbcr_format.chroma_subsampling_x, widths[1]);
+	uniform_cb_offset.y = compute_chroma_offset(
+		ycbcr_format.cb_y_position, ycbcr_format.chroma_subsampling_y, heights[1]);
+
+	uniform_cr_offset.x = compute_chroma_offset(
+		ycbcr_format.cr_x_position, ycbcr_format.chroma_subsampling_x, widths[2]);
+	uniform_cr_offset.y = compute_chroma_offset(
+		ycbcr_format.cr_y_position, ycbcr_format.chroma_subsampling_y, heights[2]);
+
 	for (unsigned channel = 0; channel < num_channels; ++channel) {
 		glActiveTexture(GL_TEXTURE0 + *sampler_num + channel);
 		check_error();
 
 		if (texture_num[channel] == 0 && (pbos[channel] != 0 || pixel_data[channel] != NULL)) {
 			GLenum format, internal_format;
-			if (channel == 1 && ycbcr_input_splitting == YCBCR_INPUT_SPLIT_Y_AND_CBCR) {
+			if (channel == 0 && ycbcr_input_splitting == YCBCR_INPUT_INTERLEAVED) {
+				if (type == GL_UNSIGNED_INT_2_10_10_10_REV) {
+					format = GL_RGBA;
+					internal_format = GL_RGB10_A2;
+				} else if (type == GL_UNSIGNED_SHORT) {
+					format = GL_RGB;
+					internal_format = GL_RGB16;
+				} else {
+					assert(type == GL_UNSIGNED_BYTE);
+					format = GL_RGB;
+					internal_format = GL_RGB8;
+				}
+			} else if (channel == 1 && ycbcr_input_splitting == YCBCR_INPUT_SPLIT_Y_AND_CBCR) {
 				format = GL_RG;
-				internal_format = GL_RG8;
+				if (type == GL_UNSIGNED_SHORT) {
+					internal_format = GL_RG16;
+				} else {
+					assert(type == GL_UNSIGNED_BYTE);
+					internal_format = GL_RG8;
+				}
 			} else {
 				format = GL_RED;
-				internal_format = GL_R8;
+				if (type == GL_UNSIGNED_SHORT) {
+					internal_format = GL_R16;
+				} else {
+					assert(type == GL_UNSIGNED_BYTE);
+					internal_format = GL_R8;
+				}
 			}
 
 			// (Re-)upload the texture.
 			texture_num[channel] = resource_pool->create_2d_texture(internal_format, widths[channel], heights[channel]);
 			glBindTexture(GL_TEXTURE_2D, texture_num[channel]);
 			check_error();
-			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
+			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, needs_mipmaps ? GL_LINEAR_MIPMAP_NEAREST : GL_LINEAR);
 			check_error();
 			glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, pbos[channel]);
 			check_error();
@@ -84,10 +131,14 @@ void YCbCrInput::set_gl_state(GLuint glsl_program_num, const string& prefix, uns
 			check_error();
 			glPixelStorei(GL_UNPACK_ROW_LENGTH, pitch[channel]);
 			check_error();
-			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, widths[channel], heights[channel], format, GL_UNSIGNED_BYTE, pixel_data[channel]);
+			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, widths[channel], heights[channel], format, type, pixel_data[channel]);
 			check_error();
 			glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
 			check_error();
+			if (needs_mipmaps) {
+				glGenerateMipmap(GL_TEXTURE_2D);
+				check_error();
+			}
 			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
 			check_error();
 			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
@@ -104,7 +155,9 @@ void YCbCrInput::set_gl_state(GLuint glsl_program_num, const string& prefix, uns
 
 	// Bind samplers.
 	uniform_tex_y = *sampler_num + 0;
-	uniform_tex_cb = *sampler_num + 1;
+	if (ycbcr_input_splitting != YCBCR_INPUT_INTERLEAVED) {
+		uniform_tex_cb = *sampler_num + 1;
+	}
 	if (ycbcr_input_splitting == YCBCR_INPUT_PLANAR) {
 		uniform_tex_cr = *sampler_num + 2;
 	}
@@ -114,40 +167,40 @@ void YCbCrInput::set_gl_state(GLuint glsl_program_num, const string& prefix, uns
 
 string YCbCrInput::output_fragment_shader()
 {
-	float offset[3];
-	Matrix3d ycbcr_to_rgb;
-	compute_ycbcr_matrix(ycbcr_format, offset, &ycbcr_to_rgb);
-
 	string frag_shader;
 
-	frag_shader = output_glsl_mat3("PREFIX(inv_ycbcr_matrix)", ycbcr_to_rgb);
-	frag_shader += output_glsl_vec3("PREFIX(offset)", offset[0], offset[1], offset[2]);
-
-	float cb_offset_x = compute_chroma_offset(
-		ycbcr_format.cb_x_position, ycbcr_format.chroma_subsampling_x, widths[1]);
-	float cb_offset_y = compute_chroma_offset(
-		ycbcr_format.cb_y_position, ycbcr_format.chroma_subsampling_y, heights[1]);
-	frag_shader += output_glsl_vec2("PREFIX(cb_offset)", cb_offset_x, cb_offset_y);
-
-	float cr_offset_x = compute_chroma_offset(
-		ycbcr_format.cr_x_position, ycbcr_format.chroma_subsampling_x, widths[2]);
-	float cr_offset_y = compute_chroma_offset(
-		ycbcr_format.cr_y_position, ycbcr_format.chroma_subsampling_y, heights[2]);
-	frag_shader += output_glsl_vec2("PREFIX(cr_offset)", cr_offset_x, cr_offset_y);
-
-	if (ycbcr_input_splitting == YCBCR_INPUT_SPLIT_Y_AND_CBCR) {
+	if (ycbcr_input_splitting == YCBCR_INPUT_INTERLEAVED) {
+		frag_shader += "#define Y_CB_CR_SAME_TEXTURE 1\n";
+	} else if (ycbcr_input_splitting == YCBCR_INPUT_SPLIT_Y_AND_CBCR) {
+		cb_cr_offsets_equal =
+			(fabs(ycbcr_format.cb_x_position - ycbcr_format.cr_x_position) < 1e-6) &&
+			(fabs(ycbcr_format.cb_y_position - ycbcr_format.cr_y_position) < 1e-6);
 		char buf[256];
-		snprintf(buf, sizeof(buf), "#define CB_CR_SAME_TEXTURE 1\n#define CB_CR_OFFSETS_EQUAL %d\n",
-			(fabs(ycbcr_format.cb_x_position - ycbcr_format.cr_x_position) < 1e-6));
+		snprintf(buf, sizeof(buf), "#define Y_CB_CR_SAME_TEXTURE 0\n#define CB_CR_SAME_TEXTURE 1\n#define CB_CR_OFFSETS_EQUAL %d\n",
+			cb_cr_offsets_equal);
 		frag_shader += buf;
 	} else {
-		frag_shader += "#define CB_CR_SAME_TEXTURE 0\n";
+		frag_shader += "#define Y_CB_CR_SAME_TEXTURE 0\n#define CB_CR_SAME_TEXTURE 0\n";
 	}
 
 	frag_shader += read_file("ycbcr_input.frag");
+	frag_shader += "#undef CB_CR_SAME_TEXTURE\n#undef Y_CB_CR_SAME_TEXTURE\n";
 	return frag_shader;
 }
 
+void YCbCrInput::change_ycbcr_format(const YCbCrFormat &ycbcr_format)
+{
+	if (cb_cr_offsets_equal) {
+		assert((fabs(ycbcr_format.cb_x_position - ycbcr_format.cr_x_position) < 1e-6) &&
+		       (fabs(ycbcr_format.cb_y_position - ycbcr_format.cr_y_position) < 1e-6));
+	}
+	if (ycbcr_input_splitting == YCBCR_INPUT_INTERLEAVED) {
+		assert(ycbcr_format.chroma_subsampling_x == 1);
+		assert(ycbcr_format.chroma_subsampling_y == 1);
+	}
+	this->ycbcr_format = ycbcr_format;
+}
+
 void YCbCrInput::invalidate_pixel_data()
 {
 	for (unsigned channel = 0; channel < 3; ++channel) {
@@ -158,8 +211,10 @@ void YCbCrInput::invalidate_pixel_data()
 bool YCbCrInput::set_int(const std::string& key, int value)
 {
 	if (key == "needs_mipmaps") {
-		// We currently do not support this.
-		return (value == 0);
+		if (ycbcr_input_splitting != YCBCR_INPUT_INTERLEAVED && value != 0) {
+			// We do not currently support this.
+			return false;
+		}
 	}
 	return Effect::set_int(key, value);
 }
